%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Misionari a kanibalove
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Autor: Pavel Zupa <pavel.zupa@gmail.com>
%
% Program hleda reseni prohledavanim stavoveho prostoru do sirky.
%
% Hlavni dotazy:
%       ?- resolve(PocetMisionaru, PocetKanibalu, 1).
%                  - najde nejkratsi reseni
%	?- allSolutions(PocetMisionaru, PocetKanibalu, 1).
%                  - vypise vsechny reseni v poradi od nejkratsiho,
%                    (tj. v poradi ve kterem se najdou)
%	?- allSolutions(3,3,1). 
%                  - vypise vsechny reseni v poradi jak se najdou 
%                    (nejkratsi bude prvni)
%
%
% s(M,K,B, N,L,C). - aktualni stav, tzn. pocet misionaru, kanibalu a lodi 
%                    na startovnim (M,K,B) a cilovem (N,L,C) brehu
%
%

:- use_module(library(lists)).
:- dynamic goal/1.

%%%%%%%%%%%%%%%%%%%%%%
%  ZAKLADNI OPERACE  %
%%%%%%%%%%%%%%%%%%%%%%

% mozne presuny osob z jednoho brehu na druhy
move(s(M,K,1,N,L,0),s(MM,K,0,NN,L,1),'M ') :-
	MM is M-1,
	NN is N+1,
	MM >= 0.
move(s(M,K,1,N,L,0),s(MM,K,0,NN,L,1),'2M') :-
	MM is M-2,
	NN is N+2,
	MM >= 0. 
move(s(M,K,1,N,L,0),s(M,KK,0,N,LL,1),'K ') :-
	KK is K-1,
	LL is L+1,
	KK >= 0.
move(s(M,K,1,N,L,0),s(M,KK,0,N,LL,1),'2K') :-
	KK is K-2,
	LL is L+2,
	KK >= 0.
move(s(M,K,1,N,L,0),s(MM,KK,0,NN,LL,1),'MK') :-
	MM is M-1,
	NN is N+1,
	KK is K-1,
	LL is L+1,
	MM >= 0,
	KK >= 0.

%%%%%%%%%%%%%%%%%%%%%%
%  POMOCNE KLAUZULE  %
%%%%%%%%%%%%%%%%%%%%%%

% udela operace move oboustranne
moveBidir(s(M,K,1,N,L,0),T,Y) :- 
	move(s(M,K,1,N,L,0), T, X), 
	atom_concat(X,'->',Y).
moveBidir(s(M,K,0,N,L,1),s(MM,KK,1,NN,LL,0),Y) :- 
	move(s(N,L,1,M,K,0), 
	s(NN,LL,0,MM,KK,1), X), 
	atom_concat(X,'<-',Y).

% kontroluje, zda na danem brehu neni nekdo poziran
validBank(M,_) :- M =:= 0.
validBank(M,K) :- M =\= 0, M >= K.

% kontroluje, zda dany stav je v poradku (nikde neni presila kanibalu)
validState(s(M,K,_,N,L,_)) :- validBank(M,K), validBank(N,L).

% provede _platny_ presun (takovy, po kterem nikdo nikoho nezere)
trans(S,T) :- moveBidir(S,T,_), validState(T).

% zobrazi pomoci symbolu prevozu cestu nalezenou pomoci bfs (seznam stavu)
printPath([_]).
printPath([A,B|Tail]) :- write(A),write(' '), moveBidir(A,B,X),write(X),write(' '),write(B),nl, printPath([B|Tail]).

% vypise pole zadanych cest
printAll([]).
printAll([Sol|Tail]) :-
	reverse(Sol, Soll), 
	printPath(Soll),
	length(Soll,L), 
	write('Length: '),
	write(L),
	nl, 
	printAll(Tail).

% samotny algoritmus prohledavani do sirky
bfs([[A|Path]|_], [A|Path]) :- 
	goal(A).

bfs([[A|Path]|Paths], Solution) :-
	bagof([N,A|Path], (
			trans(A,N), 
			non_member(N,Path)
		), NewPaths),
	append(Paths, NewPaths, NewNewPaths),!,
	bfs(NewNewPaths, Solution);
	bfs(Paths, Solution).

%%%%%%%%%%%%%%%%%%%%%
%  HLAVNI PROGRAMY  %
%%%%%%%%%%%%%%%%%%%%%

% najde reseni prechodu ze stavu Start do End 
resolve(Start, End) :- 
	retractall(goal(End)),
	asserta(goal(End)),
	bfs([[Start]], Solution), reverse(Solution, Sol), printPath(Sol), length(Sol,L), write('Length: '),write(L).

% zjednoduseni zadavani resolve(Start, End)
resolve(M,K,1) :- resolve(s(M,K,1,0,0,0), s(0,0,0,M,K,1)).

% vypise vsechny reseni
allSolutions(Start, End) :-
	retractall(goal(End)),
	asserta(goal(End)),
	bagof(Solution, bfs([[Start]], Solution), Sols),
	printAll(Sols), length(Sols, L), write('Count: '),write(L).

% zjednodusene zadani pro vsechna reseni
allSolutions(M,K,1) :- allSolutions(s(M,K,1,0,0,0), s(0,0,0,M,K,1)).


